#
#  currency.rb  --  Currency values
#

class ComputationError < StandardError ; end

class Currency

  attr_reader :cents
  alias cent cents

  class <<self
    private
    def set_res dot
      @re_valid = /[^-+(),.#{dot}_ 0-9]+/
      @re_sep   = /[,.#{dot}](\d*)\z/
      @re_spc   = /[,.#{dot}_ ]/
      not dot
    end

  end

  if RUBY_VERSION < "1.9" then
    @re_valid = /[^-+(),.\xc2\xb7_ 0-9]+/u
    @re_sep   = /[,.\xc2\xb7](\d*)\z/u
    @re_spc   = /[,.\xc2\xb7_ ]/u
  else
    set_res "\u{b7}"
  end

  class <<self

    alias cents new
    def new val = nil
      case val
        when String then
          v = val.dup
          unit = v.slice! @re_valid
          v.strip!
          case v
            when /\A\((.*)\)\z/ then
              neg = true
              v = $1.strip
            when /\b\s*-\s*\z/ then
              neg = true
              v = $`
            when /\A[+-]/ then
              neg = $& == "-"
              v = $'.lstrip
          end
          v.sub! /\B-\z/, "0"
          nk = if v =~ @re_sep and $1.length <= 2 then
            2 - $1.length
          else
            2
          end
          v << "0" * nk
          v.gsub! @re_spc, ""
          v =~ /\A\d*\z/ or
            raise ArgumentError, "Not a Currency value: \"#{val}\"."
          c = v.to_i
          c *= -1 if neg
          cents c
        when Currency then
          val.dup
        when Numeric then
          cents 100 * val
        when nil then
          zero
        else
          cents 100 * val.to_f
      end
    rescue Encoding::CompatibilityError
      retry if set_res nil
    end
    alias parse new

    def zero
      cents 0
    end

  end

  def initialize cents
    @cents = cents
    norm!
  end

  def initialize_copy oth
    @cents = oth.cents
  end

  private

  def norm!
    @cents = @cents.round.to_i
  end

  public

  def to_curr
    self
  end

  def to_f
    @cents/100.0
  end

  def to_i
    to_f.round
  end

  def succ
    Currency.cents @cents.succ
  end

  def succ!
    @cents = @cents.succ
    self
  end

  def pred
    Currency.cents @cents - 1
  end

  def pred!
    @cents -= 1
    self
  end

  def zero?
    @cents.zero?
  end

  def nonzero?
    @cents.nonzero? and self
  end

  def clear
    @cents = 0
  end

  def add oth
    @cents += oth.to_curr.cents
    self
  end

  def sub oth
    @cents -= oth.to_curr.cents
    self
  end

  def mul oth
    @cents *= oth
    norm!
    self
  end

  def div oth
    @cents /= oth
    norm!
    self
  end

  def + oth
    dup.add oth
  end

  def - oth
    dup.sub oth
  end

  def * oth
    dup.mul oth
  end

  def / oth
    case oth
      when Currency then
        begin
          Rational @cents, oth.cents
        rescue NameError
          @cents.to_f / oth.cents
        end
      else
        dup.div oth
    end
  end

  def neg!
    @cents = -@cents
    self
  end

  def +@
    self
  end

  def -@
    dup.neg!
  end

  def abs
    if @cents >= 0 then
      self
    else
      dup.neg!
    end
  end

  def neg?
    @cents < 0
  end

  def pos?
    @cents > 0
  end

  def == oth
    case oth
      when Currency then @cents == oth.cents
      when Numeric  then @cents == (oth*100).round
      when nil      then @cents.zero?
      else               oth == self
    end
  end

  def <=> oth
    case oth
      when Currency then @cents <=> oth.cents
      when Numeric  then @cents <=> (oth*100).round
      when nil      then @cents.zero? ? 0 : 1
      else               -(oth <=> self)
    end
  end
  include Comparable

  def coerce oth
    case oth
      when Currency then [ oth, self]
      when Integer  then [ (Currency.new oth), self]
      else               [ oth.to_f, to_f]
    end
  end


  @dec, @grp, @dash, @symbol = *%w(. , - $)
  class <<self
    attr_accessor :dec, :grp, :dash, :symbol
    attr_accessor :fmt, :raw_fmt
  end

  # :call-seq:
  #   cur.to_s format...   -> str
  #
  # Format currency. <code>format</code> is a hash that may contain
  # the following values:
  #
  #   :width      int        Fill left with blanks up to that width
  #   :nozero     bool       Blanks or empty string for zero value
  #   :grp        bool       Group thousands 1.234.567,00
  #   :chgrp      char       Group character (.)
  #   :neg        sym/char   negative values: "p", "(", :par = parentheses
  #                                           "r", :right    = right minus
  #   :pos        bool       Plus sign for positive values
  #   :dash       bool       Abbreviate zero decimals (100,-)
  #   :chdash     char       Dash sign
  #   :chdec      char       Separator for decimals (,)
  #   :sympos     int        symbol position: -2=left with space, -1=left, ...
  #   :symbol     char       Currency symbol
  #
  def to_s fmt = nil
    fmt ||= Currency.fmt || {}
    if zero? and fmt[ :nozero] then
      return " "*fmt[:width].to_i
    end
    c = if @cents < 0 then
      neg = true
      -@cents
    else
      @cents
    end
    s = mk_digits c, fmt
    mk_neg  s, neg, fmt
    mk_just s, fmt
    mk_sym  s, fmt
    s
  end
  alias format to_s

  private

  def mk_digits c, fmt
    v, c = c.divmod 100
    if fmt[:grp] then
      s = ""
      loop do
        v, g = v.divmod 1000
        if v.zero? then
          s.insert 0, g.to_s
          break
        else
          s.insert 0, "%03d" % g
        end
        s.insert 0, fmt[:chgrp]||Currency.grp
      end
    else
      s = v.to_s
    end
    s << (fmt[:chdec]||Currency.dec) << if fmt[:dash] and c.zero? then
      (fmt[:chdash]||Currency.dash).ljust 2
    else
      "%02d" % c
    end
  end

  def mk_neg s, neg, fmt
    case fmt[ :neg]
      when :parentheses, :par, "(", ")", /\Ap/ then
        if neg then
          s.insert 0, "("
          s << ")"
        else
          s << " "
        end
      when :right, /\Ar/ then
        s << neg ? "-" : " "
      else
        if neg then
          s.insert 0, "-"
        else
          s.insert 0, "+" if fmt[ :pos]
        end
    end
  end

  def mk_just s, fmt
    if (w = fmt[:width]) and w > s.length then
      p = s =~ /[0-9+-]/
      s.insert p, " "*(w - s.length)
    end
  end

  def mk_sym s, fmt
    y = fmt[ :sympos].to_i
    if y > 0 then
      p = s =~ /[^0-9+-]*\z/
      s.insert p, fmt[ :symbol]||Currency.symbol
      s.insert p, " " if y > 1
    elsif y < 0
      p = s =~ /[ 0-9+-]/
      s.insert p, " " if y < -1
      s.insert p, fmt[ :symbol]||Currency.symbol
    end
  end

  public

  # :call-seq:
  #   cur.raw     -> str
  #
  # Build a simple string representation without spaces, group characters,
  # parentheses etc. that can be understood by other programming language (An
  # SQL, for example).
  #
  #   c = Currency "335,00"
  #   c.raw                  #=>  "335.00"
  #
  def raw
    fmt ||= Currency.raw_fmt || {}
    c, r, dot = @cents, "", nil
    if c < 0 then
      c = -c
      s = "-"
    end
    until dot and c.zero? do
      r << (dot = fmt[ :chdec]||Currency.dec) if r.size == 2
      c, m = c.divmod 10
      r << m.to_s
    end
    r << s if s
    r.reverse!
    mk_sym r, fmt
    r
  end

  def inspect
    "#<Currency #{raw}>"
  end

  def hash
    @cents.hash
  end

end

def Currency v
  Currency.new v
end



class Percent

  attr_reader :factor

  class <<self

    def num_type
      BigDecimal
    rescue NameError
      require "bigdecimal" and retry
      raise
    end

    alias from_factor new
    def new val
      v = case val
        when Integer then num_type.new val.to_s
        when Numeric then val
        when String  then str2num val.dup
        when Percent then val.dup
        else              str2num val.to_s
      end
      v /= 100
      from_factor v
    end

    private

    def str2num str
      str.sub! ",", "."
      num_type.new str
    end

  end

  def initialize factor
    n = self.class.num_type
    @factor = case factor
      when n        then factor
      when Rational then n.new "%f" % factor
      else               n.new factor.to_s
    end
  end

  def of amount
    amount * @factor
  end

  def over amount
    amount + amount * @factor
  end

  def split g
    n, s, r = (g / (@factor + 1)), nil, nil
    loop do
      s = of n
      c = n + s
      if c < g then
        n = n.succ
        r = true
      elsif c > g then
        r and raise ComputationError, "#{g} cannot be built from #{to_s}."
        n = n.pred
      else
        break
      end
    end
    [ n, s]
  rescue NoMethodError
    raise ArgumentError,
        "Objects of class #{g.class} cannot be splitted into percentages."
  end


  def to_f
    @factor.to_f
  end

  def to_s
    b = @factor * 100
    s = if b.frac.zero? then "%d" % b else b.to_s "F" end
    s << "%"
  end

  def inspect
    "#<Percent #{to_s}>"
  end

  def coerce oth
    case oth
      when Percent
        return oth, self
      else
        return oth.to_f, to_f
    end
  end


  def + oth
    f = (@factor + 1) * (oth.factor + 1) - 1
    Percent.from_factor f
  end

  def - oth
    f = (@factor + 1) / (oth.factor + 1) - 1
    Percent.from_factor f
  end

  def +@
    self
  end

  def -@
    Percent.from_factor -@factor
  end

  def days d, leap = nil
    Percent.from_factor @factor**(d/(y leap))
  end

  def days_lin d, leap = nil
    Percent.new @factor*d/(y leap)
  end

  private

  def y leap
    leap ? 366.0 : 365.0
  end

end

def Percent p
  Percent.new p
end


class NilClass
  def to_curr
    Currency.new self
  end
  def to_perc
    Percent.new self
  end
end

class Numeric
  def to_curr
    Currency.new self
  end
  def to_perc
    Percent.new self
  end
end

class String
  def to_curr
    Currency.new self rescue Currency.zero
  end
  def to_perc
    Percent.new self
  end
end

