# encoding: UTF-8

#
#  deutsch.rb  --  Deutsche Formulierungen
#

require "date"
require "time"


class Object
  def deutsch ; to_s ; end
end


class Numeric
  def deutsch *args
    r = to_s *args
    r.gsub! /(\.)|,/ do $1 ? "," : "." end
    r
  end
end

class Float
  def deutsch format = nil
    r = format ? format % self : self.to_s
    r.gsub! /(\.)|,/ do $1 ? "," : "." end
    r
  end
end


class Time

  # :call-seq:
  #   time.deutsch   -> str
  #
  # Zeit wie im Deutschen üblicherweise angegeben.
  #
  def deutsch
    strftime "%d.%m.%Y %H:%M:%S"
  end

end


module DateDeutsch

  ABBR_DAYNAMES = %w(So Mo Di Mi Do Fr Sa)
  ABBR_MONTHNAMES = [nil] + %w(Jan Feb Mär Apr Mai Jun Jul Aug Sep Okt Nov Dez)
  DAYNAMES = %w(Sonntag Montag Dienstag Mittwoch Donnerstag Freitag Samstag)
  MONTHNAMES = [nil] + %w(Januar Februar März      April   Mai      Juni
                          Juli   August  September Oktober November Dezember)

  ABBR_DAYS = { "so"=>0, "mo"=>1, "di"=>2, "mi"=>3, "do"=>4, "fr"=>5, "sa"=>6}
  ABBR_MONTHS = { "jan"=> 1, "feb"=> 2, "m[äÄ]r"=> 3, "apr"=> 4,
                  "mai"=> 5, "jun"=> 6, "jul"   => 7, "aug"=> 8,
                  "sep"=> 9, "okt"=>10, "nov"   =>11, "dez"=>12}
  DAYS = { "sonntag"=>0, "montag"=>1, "dienstag"=>2, "mittwoch"=>3,
           "donnerstag"=>4, "freitag"=>5, "samstag"=>6}
  MONTHS = { "januar"=>1, "februar"=>2, "m[äÄ]rz"=>3, "april"=>4, "mai"=>5,
             "juni"=>6, "juli"=>7, "august"=>8, "september"=>9, "oktober"=>10,
             "november"=>11, "dezember"=>12}
  PARSE_DAYPAT   = ABBR_DAYS  .keys.join "|"
  PARSE_MONTHPAT = ABBR_MONTHS.keys.join "|"

  MONTH_CONTEXT = 9

end


class Date

  DateDeutsch.constants.each { |c| remove_const c rescue nil }
  include DateDeutsch

  class <<self
    def parse_deutsch str
      str or return
      str_n = str.dup
      self::MONTHS.each      { |m,n| str_n.gsub! /#{m}/i, "#{n}." }
      self::ABBR_MONTHS.each { |m,n| str_n.gsub! /#{m}/i, "#{n}." }
      d, m, y = str_n.split( ".", -1).map { |x| x.to_i }
      y ||= 0
      return nil unless m and m.nonzero? and d and d.nonzero?
      if y.zero? then
        n = Date.today
        y = n.year
        if m <= n.month - self::MONTH_CONTEXT then
          y += 1
        elsif m > n.month + self::MONTH_CONTEXT then
          y -= 1
        end
      elsif y < 100 then
        y += y < 50 ? 2000 : 1900
      end
      Date.new y, m, d rescue nil
    end
  end

  # :call-seq:
  #   date.deutsch   -> str
  #
  # Datum wie im Deutschen üblicherweise angegeben.
  #
  def deutsch
    strftime "%d.%m.%Y"
  end

end

class DateTime

  def deutsch
    strftime "%d.%m.%Y %H:%M:%S"
  end

end


class Array

  # :call-seq:
  #   ary.aufzaehl( mz = nil)   -> str
  #
  # Aufzählung bilden. Wenn +mz+ angegeben, dann wird dies der Mehrzahl
  # vorangestellt.
  #
  #   [ "der Rock", "die Weste", "das Hemd"].aufzaehl
  #                   #=> "Rock, Weste und Hemd"
  #   [ "der Rock", "die Weste", "das Hemd"].aufzaehl "von"
  #                   #=> "von Rock, Weste und Hemd"
  #   [ "das Hemd"].aufzaehl
  #                   #=> "das Hemd"
  #   [].aufzaehl
  #                   #=> "nichts"
  #   [].aufzaehl "von"
  #                   #=> "von nichts"
  #
  def aufzaehl mz = nil
    case length
      when 0 then
        z = "nichts"
      when 1 then
        return first
      else
        ohne = map { |e| e =~ /\A(der|die|das|des|dem|den)\s+/ ? $' : e }
        letzt = ohne.pop
        z = ([ (ohne.join ", "), letzt].join " und ")
    end
    r = ""
    r << mz << " " if mz
    r << z
  end

end


module Kernel

  # :call-seq:
  #   ja_antwort    -> match or nil
  #
  # Fragt aus der Standardeingabe nach Ja oder Nein.
  #
  def ja_antwort prompt = nil
    if prompt then
      print prompt.strip, " "
    end
    $stdout.flush
    $stdin.gets =~ /^(j|y|oui|da|evet)/i
  end
  module_function :ja_antwort

end

